#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("exoplanet").version
except DistributionNotFound:
    __version__ = "unknown version"


# General stuff
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "myst_nb",
    "IPython.sphinxext.ipython_console_highlighting",
]

myst_enable_extensions = ["dollarmath", "colon_fence"]

templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

# General information about the project.
project = "exoplanet"
author = "Dan Foreman-Mackey"
copyright = "2018, 2019, 2020, 2021, " + author

version = __version__
release = __version__

exclude_patterns = ["_build"]

# HTML theme
html_theme = "sphinx_book_theme"
html_copy_source = True
html_show_sourcelink = True
html_sourcelink_suffix = ""
html_title = "exoplanet"
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.png"
html_static_path = ["_static"]
html_css_files = ["exoplanet.css"]
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/exoplanet-dev/exoplanet",
    "repository_branch": "main",
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "notebook_interface": "classic",
    },
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
    "use_download_button": True,
}
# jupyter_execute_notebooks = "off"
nb_execution_mode = "cache"
nb_execution_timeout = -1



__all__ = [
    "LimbDarkLightCurve",
    "SecondaryEclipseLightCurve",
    "InterpolatedLightCurve",
]

from exoplanet.light_curves.interpolated import InterpolatedLightCurve
from exoplanet.light_curves.limb_dark import LimbDarkLightCurve
from exoplanet.light_curves.secondary_eclipse import SecondaryEclipseLightCurve



import logging

import numpy as np
import pytest
from scipy.stats import beta, halfnorm, kstest, rayleigh

from exoplanet.compat import USING_PYMC3, pm
from exoplanet.distributions.distributions import (
    angle,
    impact_parameter,
    quad_limb_dark,
    unit_disk,
)
from exoplanet.distributions.eccentricity import kipping13, vaneylen19


class _Base:
    random_seed = 20160911

    def _sample(self, **kwargs):
        logger = logging.getLogger("pymc3" if USING_PYMC3 else "pymc")
        logger.propagate = False
        logger.setLevel(logging.ERROR)
        kwargs["draws"] = kwargs.get("draws", 1000)
        kwargs["progressbar"] = kwargs.get("progressbar", False)
        kwargs["random_seed"] = kwargs.get("random_seed", self.random_seed)
        if USING_PYMC3:
            kwargs["return_inferencedata"] = True
            kwargs["compute_convergence_checks"] = False
        return pm.sample(**kwargs)

    def _model(self, **kwargs):
        np.random.seed(self.random_seed)
        return pm.Model(**kwargs)


class TestEccentricity(_Base):
    random_seed = 19910626

    def test_kipping13(self):
        with self._model() as model:
            kipping13("ecc", fixed=False, shape=(5, 2))
            if USING_PYMC3:
                assert "ecc_alpha" in model.named_vars
                assert "ecc_beta" in model.named_vars
            else:
                assert "ecc::alpha" in model.named_vars
                assert "ecc::beta" in model.named_vars
            trace = self._sample()

        ecc = trace.posterior["ecc"].values
        assert np.all((0 <= ecc) & (ecc <= 1))

    def test_kipping13_all(self):
        with self._model():
            kipping13("ecc", fixed=True, shape=2)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all((0 <= ecc) & (ecc <= 1))

        cdf = lambda x: beta.cdf(x, 1.12, 3.09)  # NOQA
        s, p = kstest(ecc, cdf)
        assert s < 0.05

    def test_kipping13_long(self):
        with self._model():
            kipping13("ecc", fixed=True, long=True, shape=3)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all((0 <= ecc) & (ecc <= 1))

        cdf = lambda x: beta.cdf(x, 1.12, 3.09)  # NOQA
        s, p = kstest(ecc, cdf)
        assert s < 0.05

    def test_kipping13_short(self):
        with self._model():
            kipping13("ecc", fixed=True, long=False, shape=4)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all((0 <= ecc) & (ecc <= 1))

        cdf = lambda x: beta.cdf(x, 0.697, 3.27)  # NOQA
        s, p = kstest(ecc, cdf)
        assert s < 0.05

    @pytest.mark.parametrize(
        "kwargs",
        [dict(lower=0.1), dict(upper=0.5), dict(lower=0.3, upper=0.4)],
    )
    def test_kipping13_bounds(self, kwargs):
        with self._model():
            kipping13("ecc", **kwargs)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all(
            (kwargs.get("lower", 0.0) <= ecc)
            & (ecc <= kwargs.get("upper", 1.0))
        )

    @pytest.mark.parametrize(
        "kwargs", [dict(lower=None, upper=None), dict(lower=0.2, upper=0.4)]
    )
    def test_kipping13_observed(self, kwargs):
        has_bounds = (
            kwargs.get("lower") is not None or kwargs.get("upper") is not None
        )
        with self._model() as model:
            # We want  to make sure to seed h and k inside the ecc_prior bounds
            _lower = kwargs.get("lower", 0.0) or 0.0
            _upper = kwargs.get("upper", 1.0) or 1.0
            init_ecc = 0.5 * (_lower + _upper)
            # Argument of periastron arbitrary to derive consistent h and k
            init_h = np.sqrt(init_ecc) * np.cos(np.pi / 4)
            init_k = np.sqrt(init_ecc) * np.sin(np.pi / 4)
            secosw, sesinw = unit_disk(
                "secosw", "sesinw", initval=[init_h, init_k]
            )
            ecc = pm.Deterministic("ecc", secosw**2 + sesinw**2)
            if has_bounds and USING_PYMC3:
                with pytest.raises(
                    NotImplementedError,
                    match="Passing an 'observed' eccentricity to a bounded prior is not"
                    " implemented with PyMC <= 3.",
                ):
                    ecc_prior = kipping13(
                        "ecc_prior", shape=2, observed=ecc, **kwargs
                    )
                return

            ecc_prior = kipping13("ecc_prior", shape=2, observed=ecc, **kwargs)

            # Is the prior added to the model as a potential?
            assert "ecc_prior" in model.named_vars
            assert ecc_prior in model.potentials

            # Is the prior taken into account when sampling the "posterior"?
            idata = self._sample()
            ecc_samples = idata.posterior["ecc"].values.flatten()

            if not has_bounds:
                cdf = lambda x: beta.cdf(x, 1.12, 3.09)  # NOQA
                s, p = kstest(ecc_samples, cdf)
                assert s < 0.05
            else:
                assert np.all(
                    (kwargs.get("lower", 0.0) <= ecc_samples)
                    & (ecc_samples <= kwargs.get("upper", 1.0))
                )

    @pytest.mark.parametrize("kwargs", [dict(fixed=False), dict(multi=True)])
    def test_vaneylen19(self, kwargs):
        with self._model() as model:
            vaneylen19("ecc", shape=(5, 2), **kwargs)

            if not kwargs.get("fixed", True):
                if USING_PYMC3:
                    assert "ecc_sigma_gauss" in model.named_vars
                    assert "ecc_sigma_rayleigh" in model.named_vars
                    assert "ecc_frac" in model.named_vars
                else:
                    assert "ecc::sigma_gauss" in model.named_vars
                    assert "ecc::sigma_rayleigh" in model.named_vars
                    assert "ecc::frac" in model.named_vars
            trace = self._sample()

        ecc = trace.posterior["ecc"].values
        assert np.all((0 <= ecc) & (ecc <= 1))

    def test_vaneylen19_single(self):
        with self._model():
            vaneylen19("ecc", fixed=True, multi=False, shape=2)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all((0 <= ecc) & (ecc <= 1))

        f = 0.76
        cdf = lambda x: (  # NOQA
            (1 - f) * halfnorm.cdf(x, scale=0.049)
            + f * rayleigh.cdf(x, scale=0.26)
        )
        s, p = kstest(ecc, cdf)
        assert s < 0.05

    def test_vaneylen19_multi(self):
        with self._model():
            vaneylen19("ecc", fixed=True, multi=True, shape=3)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all((0 <= ecc) & (ecc <= 1))

        f = 0.08
        cdf = lambda x: (  # NOQA
            (1 - f) * halfnorm.cdf(x, scale=0.049)
            + f * rayleigh.cdf(x, scale=0.26)
        )
        s, p = kstest(ecc, cdf)
        assert s < 0.05

    @pytest.mark.parametrize(
        "kwargs",
        [dict(lower=0.1), dict(upper=0.5), dict(lower=0.3, upper=0.4)],
    )
    def test_vaneylen19_bounds(self, kwargs):
        with self._model():
            vaneylen19("ecc", **kwargs)
            trace = self._sample()

        ecc = trace.posterior["ecc"].values.flatten()
        assert np.all(
            (kwargs.get("lower", 0.0) <= ecc)
            & (ecc <= kwargs.get("upper", 1.0))
        )

    @pytest.mark.parametrize(
        "kwargs", [dict(lower=None, upper=None), dict(lower=0.2, upper=0.4)]
    )
    def test_vaneylen19_observed(self, kwargs):
        with self._model() as model:
            # We want  to make sure to seed h and k inside the ecc_prior bounds
            _lower = kwargs.get("lower", 0.0) or 0.0
            _upper = kwargs.get("upper", 1.0) or 1.0
            init_ecc = 0.5 * (_lower + _upper)
            # Argument of periastron arbitrary to derive consistent h and k
            init_h = np.sqrt(init_ecc) * np.cos(np.pi / 4)
            init_k = np.sqrt(init_ecc) * np.sin(np.pi / 4)
            print(init_h, init_k)
            secosw, sesinw = unit_disk(
                "secosw", "sesinw", initval=[init_h, init_k]
            )
            if USING_PYMC3:
                ecc_prior = vaneylen19(
                    "ecc",
                    fixed=True,
                    multi=False,
                    observed=secosw**2 + sesinw**2,
                    **kwargs,
                )
            else:
                ecc = pm.Deterministic("ecc", secosw**2 + sesinw**2)
                ecc_prior = vaneylen19(
                    "ecc_prior",
                    fixed=True,
                    multi=False,
                    observed=ecc,
                    **kwargs,
                )

            if not USING_PYMC3:
                # Is the prior added to the model as a potential?
                assert "ecc_prior" in model.named_vars
                assert ecc_prior in model.potentials

            trace = self._sample()

        ecc_samples = trace.posterior["ecc"].values.flatten()

        if kwargs.get("lower") is None and kwargs.get("upper") is None:
            assert np.all((0 <= ecc_samples) & (ecc_samples <= 1))
            f = 0.76
            cdf = lambda x: (  # NOQA
                (1 - f) * halfnorm.cdf(x, scale=0.049)
                + f * rayleigh.cdf(x, scale=0.26)
            )
            s, p = kstest(ecc_samples, cdf)
            assert s < 0.05
        else:
            assert np.all(
                (kwargs.get("lower", 0.0) <= ecc_samples)
                & (ecc_samples <= kwargs.get("upper", 1.0))
            )


class TestUnitDisk(_Base):
    random_seed = 19930609

    def test_unit_disk(self):
        with self._model():
            h, k = unit_disk("h", "k")
            pm.Deterministic("ecc", h**2 + k**2)
            trace = self._sample()

        h_samples = trace.posterior["h"].values.flatten()
        k_samples = trace.posterior["k"].values.flatten()
        ecc_samples = trace.posterior["ecc"].values.flatten()

        # Check h and k in expected intervals
        assert np.all(h_samples < 1.0)
        assert np.all(h_samples > -1.0)
        assert np.all(k_samples < 1.0)
        assert np.all(k_samples > -1.0)

        # Check radius (eccentricity) is physical
        assert np.all(ecc_samples >= 0.0)
        assert np.all(ecc_samples < 1.0)

        # Check radius (eccentricity) is uniform
        cdf = lambda x: np.clip(x, 0, 1)  # NOQA
        s, p = kstest(ecc_samples, cdf)
        assert s < 0.05

    @pytest.mark.parametrize(
        "shape",
        [2, (1, 3), (4, 5)],
    )
    def test_unit_disk_shape(self, shape):
        shape_as_tuple = (shape,) if isinstance(shape, int) else shape
        with self._model():
            h, k = unit_disk("h", "k", shape=shape)

        if USING_PYMC3:
            assert h.tag.test_value.shape == shape_as_tuple
            assert k.tag.test_value.shape == h.tag.test_value.shape
        else:
            assert h.type.shape == shape_as_tuple
            assert k.type.shape == h.type.shape

    @pytest.mark.parametrize(
        "shape",
        [3, (4, 1), (4, 5)],
    )
    def test_unit_disk_initval(self, shape):
        shape_as_tuple = (shape,) if isinstance(shape, int) else shape

        with self._model():
            h, k = unit_disk(
                "h", "k", shape=shape, initval=np.zeros((2,) + shape_as_tuple)
            )

        if USING_PYMC3:
            assert h.tag.test_value.shape == shape_as_tuple
            assert k.tag.test_value.shape == h.tag.test_value.shape
        else:
            assert h.type.shape == shape_as_tuple
            assert k.type.shape == h.type.shape


class TestAngle(_Base):
    random_seed = 19900101

    def test_angle(self):
        with self._model():
            angle("theta")
            trace = self._sample()

        theta_samples = trace.posterior["theta"].values.flatten()

        # Check h and k in expected intervals
        assert np.all(np.abs(theta_samples) < np.pi)

        # Check theta is uniform
        cdf = lambda x: np.clip(  # NOQA
            (x + np.pi) / (2 * np.pi), -np.pi, np.pi
        )
        s, p = kstest(theta_samples, cdf)
        assert s < 0.05

    @pytest.mark.parametrize(
        "shape",
        [2, (1, 3), (4, 5)],
    )
    def test_angle_shape(self, shape):
        shape_as_tuple = (shape,) if isinstance(shape, int) else shape
        with self._model():
            theta = angle("theta", shape=shape)

        if USING_PYMC3:
            assert theta.tag.test_value.shape == shape_as_tuple
        else:
            assert theta.type.shape == shape_as_tuple

    @pytest.mark.parametrize(
        "shape",
        [3, (4, 1), (4, 5)],
    )
    def test_angle_initval(self, shape):
        shape_as_tuple = (shape,) if isinstance(shape, int) else shape
        with self._model():
            theta = angle("theta", shape=shape, initval=np.zeros(shape))

        if USING_PYMC3:
            assert theta.tag.test_value.shape == shape_as_tuple
        else:
            assert theta.type.shape == shape_as_tuple


class TestPhysical(_Base):
    random_seed = 19860925

    def test_quad_limb_dark(self):
        with self._model():
            quad_limb_dark("u")
            trace = self._sample()

        u1 = trace.posterior["u"].values[..., 0].flatten()
        u2 = trace.posterior["u"].values[..., 1].flatten()

        # Make sure that the physical constraints are satisfied
        assert np.all(u1 + u2 < 1)
        assert np.all(u1 > 0)
        assert np.all(u1 + 2 * u2 > 0)

        # Make sure that the qs are uniform
        q1 = (u1 + u2) ** 2
        q2 = 0.5 * u1 / (u1 + u2)

        cdf = lambda x: np.clip(x, 0, 1)  # NOQA
        for q in (q1, q2):
            s, p = kstest(q, cdf)
            assert s < 0.05

    def test_impact(self):
        lower = 0.1
        upper = 1.0
        shape = (5, 2)
        with self._model():
            r = pm.Uniform("r", lower=lower, upper=upper, shape=shape)
            impact_parameter("b", r, shape=shape)
            trace = self._sample()

        u = trace.posterior["r"].values
        u = np.reshape(u, u.shape[:2] + (-1,))
        cdf = lambda x: np.clip((x - lower) / (upper - lower), 0, 1)  # NOQA
        for i in range(u.shape[-1]):
            s, p = kstest(u[..., i].flatten(), cdf)
            assert s < 0.05

        assert np.all(
            trace.posterior["b"].values <= 1 + trace.posterior["r"].values
        )

    @pytest.mark.skipif(
        USING_PYMC3, reason="Automatic shape inference doesn't work in PyMC3"
    )
    def test_impact_shape_inference(self):
        lower = 0.1
        upper = 1.0
        shape = (5, 2)
        with self._model():
            r = pm.Uniform("r", lower=lower, upper=upper, shape=shape)
            impact_parameter("b", r)
            trace = self._sample()

        assert trace.posterior["b"].values.shape[-2:] == shape

        u = trace.posterior["r"].values
        u = np.reshape(u, u.shape[:2] + (-1,))
        cdf = lambda x: np.clip((x - lower) / (upper - lower), 0, 1)  # NOQA
        for i in range(u.shape[-1]):
            s, p = kstest(u[..., i].flatten(), cdf)
            assert s < 0.05

        assert np.all(
            trace.posterior["b"].values <= 1 + trace.posterior["r"].values
        )



import os

if os.environ.get("EXOPLANET_FORCE_PYMC3", "0") == "1":
    import pymc3 as pm

    USING_PYMC3 = True

else:
    try:
        import pymc as pm

        USING_PYMC3 = False

    except ImportError:
        try:
            import pymc3 as pm

            USING_PYMC3 = True

        except ImportError:
            raise ImportError(
                "No compatible version of PyMC found. "
                "Please install either pymc3 or pymc>=4."
            )

if USING_PYMC3:
    from exoplanet_core.pymc3 import ops as ops
    from theano import function as function
    from theano import grad as grad
    from theano import tensor as tensor
    from theano.ifelse import ifelse as ifelse

    tensor.abs = tensor.abs_

    try:
        from theano.assert_op import Assert as Assert
    except ImportError:
        from theano.tensor.opt import Assert as Assert

    from theano.gradient import verify_grad as verify_grad

    try:
        import theano

        change_flags = theano.config.change_flags
    except (ImportError, AttributeError):
        from theano.configparser import change_flags as change_flags

else:
    import pytensor
    from pytensor import function as function
    from pytensor import grad as grad
    from pytensor import tensor as tensor

    change_flags = pytensor.config.change_flags
    from exoplanet_core.pymc import ops as ops
    from pytensor.gradient import verify_grad as verify_grad
    from pytensor.ifelse import ifelse as ifelse
    from pytensor.raise_op import Assert as Assert



__all__ = [
    "ImpactParameter",
    "QuadLimbDark",
    "angle",
    "unit_disk",
    "impact_parameter",
    "quad_limb_dark",
    "eccentricity",
]

from exoplanet.distributions import eccentricity
from exoplanet.distributions.distributions import (
    angle,
    impact_parameter,
    quad_limb_dark,
    unit_disk,
)
from exoplanet.utils import deprecated

# For backwards compatibility, define wrappers for the old Distribution-based
# versions of these functions
ImpactParameter = deprecated(
    alternate="exoplanet.distributions.impact_parameter"
)(impact_parameter)
QuadLimbDark = deprecated(alternate="exoplanet.distributions.quad_limb_dark")(
    quad_limb_dark
)



__all__ = ["LimbDarkLightCurve"]

import numpy as np

from exoplanet.citations import add_citations_to_model
from exoplanet.compat import Assert, ops
from exoplanet.compat import tensor as pt
from exoplanet.utils import as_tensor_variable, deprecation_warning


def get_cl(u1, u2):
    u1 = as_tensor_variable(u1)
    u2 = as_tensor_variable(u2)
    c0 = 1 - u1 - 1.5 * u2
    c1 = u1 + 2 * u2
    c2 = -0.25 * u2
    norm = np.pi * (c0 + c1 / 1.5)
    return pt.stack([c0, c1, c2]) / norm


def quad_limbdark_light_curve(c, b, r):
    b = as_tensor_variable(b)
    r = as_tensor_variable(r)
    return pt.dot(ops.quad_solution_vector(b, r), c) - 1.0


class LimbDarkLightCurve:
    """A quadratically limb darkened light curve

    .. note:: Earlier versions of exoplanet supported higher (and lower) order
        limb darkening, but this support was removed in v0.5.0. For higher
        order limb darkening profiles, use starry directly.

    Args:
        u1 (scalar): The first limb darkening coefficient
        u2 (scalar): The second limb darkening coefficient
    """

    __citations__ = ("starry",)

    def __init__(self, u1, u2=None, model=None):
        add_citations_to_model(self.__citations__, model=model)
        if u2 is None:
            deprecation_warning(
                "using a vector of limb darkening coefficients is deprecated; "
                "use u1 and u2 directly"
            )

            # If a vector is provided, we need to assert that it is 2D
            msg = (
                "only quadratic limb darkening is supported; "
                "use `starry` for more flexibility"
            )
            try:
                assert_op = Assert(msg)
            except AttributeError:
                assert_op = Assert(msg)
            u1 = as_tensor_variable(u1)
            u1 = assert_op(
                u1, pt.and_(pt.eq(u1.ndim, 1), pt.eq(u1.shape[0], 2))
            )

            self.u1 = u1[0]
            self.u2 = u1[1]
        else:
            self.u1 = as_tensor_variable(u1)
            self.u2 = as_tensor_variable(u2)
        self.c = get_cl(self.u1, self.u2)

    def get_ror_from_approx_transit_depth(self, delta, b, jac=False):
        """Get the radius ratio corresponding to a particular transit depth

        This result will be approximate and it requires ``|b| < 1`` because it
        relies on the small planet approximation.

        Args:
            delta (tensor): The approximate transit depth in relative units
            b (tensor): The impact parameter
            jac (bool): If true, the Jacobian ``d ror / d delta`` is also
                returned

        Returns:
            ror: The radius ratio that approximately corresponds to the depth
            ``delta`` at impact parameter ``b``.

        """
        b = as_tensor_variable(b)
        delta = as_tensor_variable(delta)
        f0 = 1 - 2 * self.u1 / 6.0 - 2 * self.u2 / 12.0
        arg = 1 - pt.sqrt(1 - b**2)
        f = 1 - self.u1 * arg - self.u2 * arg**2
        factor = f0 / f
        ror = pt.sqrt(delta * factor)
        if not jac:
            return pt.reshape(ror, b.shape)
        drorddelta = 0.5 * factor / ror
        return pt.reshape(ror, b.shape), pt.reshape(drorddelta, b.shape)

    def get_light_curve(
        self,
        orbit=None,
        r=None,
        t=None,
        texp=None,
        oversample=7,
        order=0,
        use_in_transit=None,
        light_delay=False,
    ):
        """Get the light curve for an orbit at a set of times

        Args:
            orbit: An object with a ``get_relative_position`` method that
                takes a tensor of times and returns a list of Cartesian
                coordinates of a set of bodies relative to the central source.
                This method should return three tensors (one for each
                coordinate dimension) and each tensor should have the shape
                ``append(t.shape, r.shape)`` or ``append(t.shape, oversample,
                r.shape)`` when ``texp`` is given. The first two coordinate
                dimensions are treated as being in the plane of the sky and the
                third coordinate is the line of sight with positive values
                pointing *away* from the observer. For an example, take a look
                at :class:`orbits.KeplerianOrbit`.
            r (tensor): The radius of the transiting body in the same units as
                ``r_star``. This should have a shape that is consistent with
                the coordinates returned by ``orbit``. In general, this means
                that it should probably be a scalar or a vector with one entry
                for each body in ``orbit``. Note that this is a different
                quantity than the planet-to-star radius ratio; do not confuse
                the two!
            t (tensor): The times where the light curve should be evaluated.
            texp (Optional[tensor]): The exposure time of each observation.
                This can be a scalar or a tensor with the same shape as ``t``.
                If ``texp`` is provided, ``t`` is assumed to indicate the
                timestamp at the *middle* of an exposure of length ``texp``.
            oversample (Optional[int]): The number of function evaluations to
                use when numerically integrating the exposure time.
            order (Optional[int]): The order of the numerical integration
                scheme. This must be one of the following: ``0`` for a
                centered Riemann sum (equivalent to the "resampling" procedure
                suggested by Kipping 2010), ``1`` for the trapezoid rule, or
                ``2`` for Simpson's rule.
            use_in_transit (Optional[bool]): If ``True``, the model will only
                be evaluated for the data points expected to be in transit
                as computed using the ``in_transit`` method on ``orbit``.

        """
        if orbit is None:
            raise ValueError("missing required argument 'orbit'")
        if r is None:
            raise ValueError("missing required argument 'r'")
        if t is None:
            raise ValueError("missing required argument 't'")

        use_in_transit = (
            not light_delay if use_in_transit is None else use_in_transit
        )

        r = as_tensor_variable(r)
        r = pt.reshape(r, (r.size,))
        t = as_tensor_variable(t)

        # If use_in_transit, we should only evaluate the model at times where
        # at least one planet is transiting
        if use_in_transit:
            transit_model = pt.shape_padleft(
                pt.zeros_like(r), t.ndim
            ) + pt.shape_padright(pt.zeros_like(t), r.ndim)
            inds = orbit.in_transit(t, r=r, texp=texp, light_delay=light_delay)
            t = t[inds]

        # Handle exposure time integration
        if texp is None:
            tgrid = t
            rgrid = pt.shape_padleft(r, tgrid.ndim) + pt.shape_padright(
                pt.zeros_like(tgrid), r.ndim
            )
        else:
            texp = as_tensor_variable(texp)

            oversample = int(oversample)
            oversample += 1 - oversample % 2
            stencil = np.ones(oversample)

            # Construct the exposure time integration stencil
            if order == 0:
                dt = np.linspace(-0.5, 0.5, 2 * oversample + 1)[1:-1:2]
            elif order == 1:
                dt = np.linspace(-0.5, 0.5, oversample)
                stencil[1:-1] = 2
            elif order == 2:
                dt = np.linspace(-0.5, 0.5, oversample)
                stencil[1:-1:2] = 4
                stencil[2:-1:2] = 2
            else:
                raise ValueError("order must be <= 2")
            stencil /= np.sum(stencil)

            if texp.ndim == 0:
                dt = texp * dt
            else:
                if use_in_transit:
                    dt = pt.shape_padright(texp[inds]) * dt
                else:
                    dt = pt.shape_padright(texp) * dt
            tgrid = pt.shape_padright(t) + dt

            # Madness to get the shapes to work out...
            rgrid = pt.shape_padleft(r, tgrid.ndim) + pt.shape_padright(
                pt.zeros_like(tgrid), 1
            )

        # Evalute the coordinates of the transiting body in the plane of the
        # sky
        coords = orbit.get_relative_position(tgrid, light_delay=light_delay)
        b = pt.sqrt(coords[0] ** 2 + coords[1] ** 2)
        b = pt.reshape(b, rgrid.shape)
        los = pt.reshape(coords[2], rgrid.shape)

        lc = self._compute_light_curve(
            b / orbit.r_star, rgrid / orbit.r_star, los / orbit.r_star
        )

        if texp is not None:
            stencil = pt.shape_padright(pt.shape_padleft(stencil, t.ndim), 1)
            lc = pt.sum(stencil * lc, axis=t.ndim)

        if use_in_transit:
            transit_model = pt.set_subtensor(transit_model[inds], lc)
            return transit_model
        else:
            return lc

    def _compute_light_curve(self, b, r, los=None):
        """Compute the light curve for a set of impact parameters and radii

        .. note:: The stellar radius is *not* included in this method so the
            coordinates should be in units of the star's radius.

        Args:
            b (tensor): A tensor of impact parameter values.
            r (tensor): A tensor of radius ratios with the same shape as ``b``.
            los (Optional[tensor]): The coordinates of the body along the
                line-of-sight. If ``los > 0`` the body is between the observer
                and the source.

        """
        b = as_tensor_variable(b)
        if los is None:
            los = pt.ones_like(b)
        lc = quad_limbdark_light_curve(self.c, b, r)
        return pt.switch(pt.gt(los, 0), lc, pt.zeros_like(lc))



import astropy.units as u
import numpy as np

from exoplanet import units
from exoplanet.orbits import KeplerianOrbit


def test_mass_units():
    P_earth = 365.256
    Tper_earth = 2454115.5208333
    inclination_earth = np.radians(45.0)

    orbit1 = KeplerianOrbit(
        period=P_earth,
        t_periastron=Tper_earth,
        incl=inclination_earth,
        m_planet=units.with_unit(1.0, u.M_earth),
    )
    orbit2 = KeplerianOrbit(
        period=P_earth,
        t_periastron=Tper_earth,
        incl=inclination_earth,
        m_planet=1.0,
        m_planet_units=u.M_earth,
    )

    t = np.linspace(Tper_earth, Tper_earth + 1000, 1000)
    rv1 = orbit1.get_radial_velocity(t).eval()
    rv_diff = np.max(rv1) - np.min(rv1)
    assert rv_diff < 1.0, "with_unit"

    rv2 = orbit2.get_radial_velocity(t).eval()
    rv_diff = np.max(rv2) - np.min(rv2)
    assert rv_diff < 1.0, "m_planet_units"
    np.testing.assert_allclose(rv2, rv1)



__all__ = ["kipping13", "vaneylen19"]

import numpy as np

from exoplanet.citations import add_citations_to_model
from exoplanet.compat import USING_PYMC3, pm
from exoplanet.compat import tensor as pt


def kipping13(
    name, fixed=True, long=None, lower=None, upper=None, model=None, **kwargs
):
    """The beta distribution parameters fit by `Kipping (2013b)
    <https://arxiv.org/abs/1306.4982>`_.

    Args:
        name (str): The name of the eccentricity variable.
        fixed (bool, optional): If ``True``, use the posterior median
            hyperparameters. Otherwise, marginalize over the parameters.
        long (bool, optional): If ``True``, use the parameters for the long
            period fit. If ``False``, use the parameters for the short period
            fit. If not given, the parameters fit using the full dataset are
            used.
        lower (float, optional): Restrict the eccentricity to be larger than
            this value.
        upper (float, optional): Restrict the eccentricity to be smaller than
            this value.

    Returns:
        The eccentricity distribution.

    """
    model = pm.modelcontext(model)
    add_citations_to_model(["kipping13b"], model=model)

    if long is None:
        # If 'long' is not provided, use the fit for the parameters from the
        # full dataset
        alpha_mu = 1.12
        alpha_sd = 0.1
        beta_mu = 3.09
        beta_sd = 0.3
    else:
        # If 'long' is set, select either the long or short period model
        # parameters
        if long:
            alpha_mu = 1.12
            alpha_sd = 0.1
            beta_mu = 3.09
            beta_sd = 0.3
        else:
            alpha_mu = 0.697
            alpha_sd = 0.4
            beta_mu = 3.27
            beta_sd = 0.3

    with model:
        if fixed:
            # Use the posterior median parameters
            alpha = alpha_mu
            beta = beta_mu
        else:
            # Marginalize over the uncertainty on the parameters of the beta
            with pm.Model(name=name):
                alpha = _truncate(
                    "alpha",
                    pm.Normal,
                    lower=0,
                    mu=alpha_mu,
                    sigma=alpha_sd,
                    **_with_initval(initval=alpha_mu),
                )
                beta = _truncate(
                    "beta",
                    pm.Normal,
                    lower=0,
                    mu=beta_mu,
                    sigma=beta_sd,
                    **_with_initval(initval=beta_mu),
                )

        # Allow for upper and lower bounds
        ecc = kwargs.pop("observed", None)
        if lower is not None or upper is not None:
            lower = 0.0 if lower is None else lower
            upper = 1.0 if upper is None else upper
            if ecc is None:
                kwargs["initval"] = kwargs.pop(
                    "initval", kwargs.pop("testval", 0.5 * (lower + upper))
                )
                return _truncate(
                    name,
                    pm.Beta,
                    lower=lower,
                    upper=upper,
                    alpha=alpha,
                    beta=beta,
                    **_with_initval(**kwargs),
                )
            else:
                if USING_PYMC3:
                    raise NotImplementedError(
                        "Passing an 'observed' eccentricity to a bounded prior is not "
                        "implemented with PyMC <= 3."
                    )
                else:
                    dist = _truncate_dist(
                        pm.Beta,
                        lower=lower,
                        upper=upper,
                        alpha=alpha,
                        beta=beta,
                        # **_with_initval(**kwargs),
                        **kwargs,
                    )
        else:
            if ecc is None:
                return pm.Beta(name, alpha=alpha, beta=beta, **kwargs)
            else:
                dist = pm.Beta.dist(alpha=alpha, beta=beta, **kwargs)

        return pm.Potential(name, _logp(dist, ecc))


def vaneylen19(
    name,
    fixed=True,
    multi=False,
    lower=None,
    upper=None,
    model=None,
    **kwargs,
):
    """The eccentricity distribution for small planets

    The mixture distribution fit by `Van Eylen et al. (2019)
    <https://arxiv.org/abs/1807.00549>`_ to a population of well-characterized
    small transiting planets observed by Kepler.

    Args:
        name (str): The name of the eccentricity variable.
        fixed (bool, optional): If ``True``, use the posterior median
            hyperparameters. Otherwise, marginalize over the parameters.
        multi (bool, optional): If ``True``, use the distribution for systems
            with multiple transiting planets. If ``False`` (default), use the
            distribution for systems with only one detected transiting planet.
        lower (float, optional): Restrict the eccentricity to be larger than
            this value.
        upper (float, optional): Restrict the eccentricity to be smaller than
            this value.

    Returns:
        The eccentricity distribution.

    """

    model = pm.modelcontext(model)
    add_citations_to_model(["vaneylen19"], model=model)

    sigma_gauss_mu = 0.049
    sigma_gauss_sd = 0.02
    sigma_rayleigh_mu = 0.26
    sigma_rayleigh_sd = 0.05
    if multi:
        frac_mu = 0.08
        frac_sd = 0.08
    else:
        frac_mu = 0.76
        frac_sd = 0.2

    with model:
        ecc = kwargs.pop("observed", None)
        _lower = 0.0 if lower is None else lower
        _upper = 1.0 if upper is None else upper
        if ecc is None or USING_PYMC3:
            ecc = pm.Uniform(
                name,
                lower=_lower,
                upper=_upper,
                **kwargs,
            )
            ecc_prior = ecc
        else:
            # TODO: Still define uniform here or put it in a single potential with other distributions below?
            # TODO: Would there be any advantage to remove this unnecessary uniform when ecc already constrained 0 < e < 1 by, for e.g., unit_disk?
            unif = pm.Uniform.dist(lower=_lower, upper=_upper, **kwargs)
            ecc_prior = pm.Potential(
                name,
                _logp(unif, ecc),
            )

        with pm.Model(name=name):
            if fixed:
                sigma_gauss = sigma_gauss_mu
                sigma_rayleigh = sigma_rayleigh_mu
                frac = frac_mu
            else:
                sigma_gauss = _truncate(
                    "sigma_gauss",
                    pm.Normal,
                    lower=0,
                    mu=sigma_gauss_mu,
                    sigma=sigma_gauss_sd,
                    **_with_initval(initval=sigma_gauss_mu),
                )
                sigma_rayleigh = _truncate(
                    "sigma_rayleigh",
                    pm.Normal,
                    lower=0,
                    mu=sigma_rayleigh_mu,
                    sigma=sigma_rayleigh_sd,
                    **_with_initval(initval=sigma_rayleigh_mu),
                )
                frac = _truncate(
                    "frac",
                    pm.Normal,
                    lower=0,
                    upper=1,
                    mu=frac_mu,
                    sigma=frac_sd,
                    **_with_initval(initval=frac_mu),
                )

            gauss = pm.HalfNormal.dist(sigma=sigma_gauss)
            rayleigh = pm.Weibull.dist(
                alpha=2, beta=np.sqrt(2) * sigma_rayleigh
            )

            pm.Potential(
                "prior",
                pm.math.logaddexp(
                    pt.log(1 - frac) + _logp(gauss, ecc),
                    pt.log(frac) + _logp(rayleigh, ecc),
                ),
            )

        return ecc_prior


def _with_initval(**kwargs):
    val = kwargs.pop("initval", kwargs.pop("testval", None))
    if USING_PYMC3:
        return dict(kwargs, testval=val)
    return dict(kwargs, initval=val)


def _logp(rv, x):
    if USING_PYMC3:
        return rv.logp(x)
    return pm.logp(rv, x)


def _truncate(name, dist, *, lower=None, upper=None, **kwargs):
    if USING_PYMC3:
        return pm.Bound(dist, lower=lower, upper=upper)(
            name, **_with_initval(**kwargs)
        )
    initval = kwargs.pop("initval", kwargs.pop("testval", None))
    return pm.Truncated(
        name, dist.dist(**kwargs), lower=lower, upper=upper, initval=initval
    )


def _truncate_dist(dist, *, lower=None, upper=None, **kwargs):
    if USING_PYMC3:
        raise NotImplementedError(
            "_truncate_dist not implemented for PyMC 3 yet"
        )
    return pm.Truncated.dist(dist.dist(**kwargs), lower=lower, upper=upper)



from exoplanet import interp as interp
from exoplanet import orbits as orbits
from exoplanet.citations import CITATIONS as CITATIONS
from exoplanet.distributions import *  # NOQA
from exoplanet.estimators import *  # NOQA
from exoplanet.exoplanet_version import __version__ as __version__
from exoplanet.light_curves import *  # NOQA
from exoplanet.utils import *  # NOQA

__bibtex__ = __citation__ = CITATIONS["exoplanet"][1]
__uri__ = "https://docs.exoplanet.codes"
__author__ = "Daniel Foreman-Mackey"
__email__ = "foreman.mackey@gmail.com"
__license__ = "MIT"
__description__ = "Fast and scalable MCMC for all your exoplanet needs"



__all__ = [
    "KeplerianOrbit",
    "get_true_anomaly",
    "TTVOrbit",
    "SimpleTransitOrbit",
    "duration_to_eccentricity",
]

from exoplanet.orbits.dur_to_ecc import duration_to_eccentricity
from exoplanet.orbits.keplerian import KeplerianOrbit, get_true_anomaly
from exoplanet.orbits.simple import SimpleTransitOrbit
from exoplanet.orbits.ttv import TTVOrbit



import tempfile
from pathlib import Path

import nox

ALL_PYTHON_VS = ["3.9", "3.10", "3.11"]


@nox.session(python=ALL_PYTHON_VS)
def test_pymc3(session):
    session.install(".[test,pymc3]")
    session.install("batman-package", "starry>=0.9.9", "tqdm")
    session.run("python", "-c", "import theano")
    session.run("python", "-c", "import exoplanet_core.pymc3.ops")
    session.run("python", "-c", "import starry")
    session.run("pytest", "-v", *session.posargs)


@nox.session(python=ALL_PYTHON_VS)
def test_pymc(session):
    session.install(".[test,pymc]")
    session.install("batman-package")
    session.run("python", "-c", "import pytensor")
    session.run("python", "-c", "import exoplanet_core.pymc.ops")
    session.run("pytest", "-v", *session.posargs)


@nox.session(python=ALL_PYTHON_VS)
def lint(session):
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session
def docs(session):
    session.install(".[docs]")
    session.install("sphinx")
    session.run(
        "sphinx-build",
        "--color",
        "-W",
        "-b",
        "dirhtml",
        "-d",
        str(session.cache_dir / "doctrees"),
        "docs",
        str(session.cache_dir / "dirhtml"),
        *session.posargs
    )
    session.run("open", session.cache_dir)



__all__ = ["angle", "unit_disk", "quad_limb_dark", "impact_parameter"]

import numpy as np

from exoplanet.citations import add_citations_to_model
from exoplanet.compat import USING_PYMC3, pm
from exoplanet.compat import tensor as pt
from exoplanet.utils import as_tensor_variable


def angle(name, *, regularization=10.0, **kwargs):
    """An angle constrained to be in the range -pi to pi

    The actual sampling is performed in the two dimensional vector space
    proportional to ``(sin(theta), cos(theta))`` so that the sampler doesn't see
    a discontinuity at pi.

    The regularization parameter can be used to improve sampling performance
    when the value of the angle is well constrained. It removes prior mass near
    the origin in the sampling space, which can lead to bad geometry when the
    angle is poorly constrained, but better performance when it is. The default
    value of ``10.0`` is a good starting point.
    """
    shape_arg = kwargs.get("shape", ())
    initval = kwargs.pop("initval", kwargs.pop("testval", np.zeros(shape_arg)))
    x1 = pm.Normal(
        f"__{name}_angle1", **_with_initval(initval=np.sin(initval), **kwargs)
    )
    x2 = pm.Normal(
        f"__{name}_angle2", **_with_initval(initval=np.cos(initval), **kwargs)
    )
    if regularization is not None:
        pm.Potential(
            f"{name}_regularization",
            regularization * pt.log(x1**2 + x2**2),
        )
    return pm.Deterministic(name, pt.arctan2(x1, x2))


def unit_disk(name_x, name_y, **kwargs):
    """Two dimensional parameters constrained to live within the unit disk

    This returns two distributions whose sum of squares will be in the range
    ``[0, 1)``. For example, in this code block:

    .. code-block:: python

        x, y = unit_disk("x", "y") radius_sq = x**2 + y**2

    the tensor ``radius_sq`` will always have a value in the range ``[0, 1)``.

    Args:
        name_x: The name of the first distribution.
        name_y: The name of the second distribution.
    """
    shape_kwarg = kwargs.get("shape", ())
    if isinstance(shape_kwarg, int):
        shape_kwarg = (shape_kwarg,)
    init_shape = (2,) + shape_kwarg
    initval = kwargs.pop(
        "initval", kwargs.pop("testval", np.zeros(init_shape))
    )
    # initval = kwargs.pop("initval", kwargs.pop("testval", [0.0, 0.0]))
    kwargs["lower"] = -1.0
    kwargs["upper"] = 1.0
    x1 = pm.Uniform(name_x, **_with_initval(initval=initval[0], **kwargs))
    x2 = pm.Uniform(
        f"__{name_y}_unit_disk",
        **_with_initval(
            initval=initval[1] * np.sqrt(1 - initval[0] ** 2), **kwargs
        ),
    )
    norm = pt.sqrt(1 - x1**2)
    pm.Potential(f"{name_y}_jacobian", pt.log(norm))
    return x1, pm.Deterministic(name_y, x2 * norm)


def quad_limb_dark(name, **kwargs):
    """An uninformative prior for quadratic limb darkening parameters

    This is an implementation of the `Kipping (2013)
    <https://arxiv.org/abs/1308.0009>`_ reparameterization of the two-parameter
    limb darkening model to allow for efficient and uninformative sampling.
    """
    add_citations_to_model(("kipping13",), kwargs.get("model", None))

    u = kwargs.pop("initval", kwargs.pop("testval", [np.sqrt(0.5), 0.0]))
    u1 = u[0]
    u2 = u[1]
    kwargs["lower"] = 0.0
    kwargs["upper"] = 1.0
    q1 = pm.Uniform(
        f"__{name}_q1", **_with_initval(initval=(u1 + u2) ** 2, **kwargs)
    )
    q2 = pm.Uniform(
        f"__{name}_q2", **_with_initval(initval=0.5 * u1 / (u1 + u2), **kwargs)
    )
    sqrtq1 = pt.sqrt(q1)
    twoq2 = 2 * q2
    return pm.Deterministic(
        name, pt.stack([sqrtq1 * twoq2, sqrtq1 * (1 - twoq2)], axis=0)
    )


def impact_parameter(name, ror, **kwargs):
    """The impact parameter distribution for a transiting planet

    Args:
        ror: A scalar, tensor, or PyMC (3 or 5+) distribution representing the radius
            ratio between the planet and star. Conditioned on a value of
            ``ror``, this will be uniformly distributed between ``0`` and
            ``1+ror``.
    """
    ror = as_tensor_variable(ror)
    bhat = kwargs.pop("initval", kwargs.pop("testval", 0.5))
    if not USING_PYMC3:
        shape = kwargs.setdefault("shape", ror.shape)
        if isinstance(shape, int):
            shape = (shape,)
        bhat = pt.broadcast_to(bhat, shape)
    kwargs["lower"] = 0.0
    kwargs["upper"] = 1.0
    norm = pm.Uniform(
        f"__{name}_impact_parameter",
        **_with_initval(initval=bhat / (1 + ror), **kwargs),
    )
    return pm.Deterministic(name, norm * (1 + ror))


def _with_initval(**kwargs):
    val = kwargs.pop("initval", kwargs.pop("testval", None))
    if USING_PYMC3:
        return dict(kwargs, testval=val)
    return dict(kwargs, initval=val)
